/* EXTNDDRV.C

 Copyright (C) 1997 Prasad Dabak and Sandeep Phadke

 Device driver implementation for adding new system services
 to the NT kernel
*/

#include "ntddk.h"
#include "stdarg.h"
#include "stdio.h"
#include "extnddrv.h"

#pragma pack(1)

#ifdef DBG
#define ExtndDrvKdPrint(arg) DbgPrint arg
#else
#define ExtndDrvKdPrint(arg)
#endif

void DebugOut(const char *fmt, ...)
{
	static va_list args;
	static char buf[1000];
	va_start(args, fmt);
	vsprintf(buf, fmt, args);
	va_end(args);
	ExtndDrvKdPrint ((buf));
	return;
}

NTSTATUS SampleService0(void)
{
	DebugOut("Kernel Service without any parameters called\n");
	_asm {
		pushad
		mov ax, 1000 
		mov bx, 200
		mov cx, ax
		mov al, 0b6h
		out 43h, al
		mov dx, 0012h
		mov ax, 34dch
		div cx
		out 42h, al
		mov al, ah
		out 42h, al
		in al, 61h
		mov ah, al
		or al, 03h
		out 61h, al
		l1:
			mov ecx, 4680
		l2:
			loop l2
			dec bx
			jnz l1
		mov al, ah
		out 61h, al
		popad
	}
	return STATUS_SUCCESS;
}

NTSTATUS SampleService1(unsigned int param)
{
	DebugOut("Kernel Service with one parameter called\n");
	return STATUS_SUCCESS+1;
}

NTSTATUS SampleService2(unsigned int param1, unsigned int param2)
{
	DebugOut("Kernel Service with two parameters called\n");
	return STATUS_SUCCESS+2;
}

NTSTATUS SampleService3(unsigned int param1, unsigned int param2,
			unsigned int param3)
{
	DebugOut("Kernel Service with three parameters called\n");
	return STATUS_SUCCESS+3;
}


NTSTATUS SampleService4(unsigned int param1, unsigned int param2,
			unsigned int param3, unsigned int param4)
{
	DebugOut("Kernel Service with four parameters called\n");
	return STATUS_SUCCESS+4;
}

NTSTATUS SampleService5(unsigned int param1, unsigned int param2,
			unsigned int param3, unsigned int param4,
			unsigned int param5)
{
	DebugOut("Kernel Service with five parameters called\n");
	return STATUS_SUCCESS+5;
}

NTSTATUS SampleService6(unsigned int param1, unsigned int param2,
			unsigned int param3, unsigned int param4,
			unsigned int param5, unsigned int param6)
{
	DebugOut("Kernel Service with six parameters called\n");
	return STATUS_SUCCESS+6;
}

NTSTATUS
ExtndDrvDispatch(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    );

VOID
ExtndDrvUnload(
    IN PDRIVER_OBJECT DriverObject
    );


NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
    PDEVICE_OBJECT         deviceObject        = NULL;
    NTSTATUS               ntStatus;
    WCHAR                  deviceNameBuffer[]  = L"\\Device\\EXTNDDRV";
    UNICODE_STRING         deviceNameUnicodeString;
    PDEVICE_EXTENSION      deviceExtension;
    WCHAR                  deviceLinkBuffer[]  = L"\\DosDevices\\EXTNDDRV";
    UNICODE_STRING         deviceLinkUnicodeString;

    ExtndDrvKdPrint (("ExtndDRV.SYS: entering DriverEntry\n"));

    RtlInitUnicodeString (&deviceNameUnicodeString,
			  deviceNameBuffer
			  );

    ntStatus = IoCreateDevice (DriverObject,
			       sizeof (DEVICE_EXTENSION),
			       &deviceNameUnicodeString,
			       FILE_DEVICE_EXTNDDRV,
			       0,
			       TRUE,
			       &deviceObject
			       );

    if (NT_SUCCESS(ntStatus))
    {
	deviceExtension = (PDEVICE_EXTENSION) deviceObject->DeviceExtension;
	RtlInitUnicodeString (&deviceLinkUnicodeString,
			      deviceLinkBuffer
			      );

	ntStatus = IoCreateSymbolicLink (&deviceLinkUnicodeString,
					 &deviceNameUnicodeString
					 );


	if (!NT_SUCCESS(ntStatus))
	{
	    ExtndDrvKdPrint (("EXTNDDRV.SYS: IoCreateSymbolicLink failed\n"));
	}
	DriverObject->MajorFunction[IRP_MJ_CREATE]         =
	DriverObject->MajorFunction[IRP_MJ_CLOSE]          =
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = ExtndDrvDispatch;
	/*DriverObject->DriverUnload                         = ExtndDrvUnload;*/
    }


done_DriverEntry:

    if (!NT_SUCCESS(ntStatus))
    {
	if (deviceObject)

	    IoDeleteDevice (deviceObject);
    }
    ntStatus=AddServices();
    return ntStatus;
}

NTSTATUS
ExtndDrvDispatch(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp
    )
{

    PIO_STACK_LOCATION  irpStack;
    PDEVICE_EXTENSION   deviceExtension;
    PVOID               ioBuffer;
    ULONG               inputBufferLength;
    ULONG               outputBufferLength;
    NTSTATUS            ntStatus;

    Irp->IoStatus.Status      = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    irpStack = IoGetCurrentIrpStackLocation (Irp);

    deviceExtension = DeviceObject->DeviceExtension;


    switch (irpStack->MajorFunction)
    {
    case IRP_MJ_CREATE:
	ExtndDrvKdPrint (("EXTNDDRV.SYS: IRP_MJ_CREATE\n"));
	break;

    case IRP_MJ_CLOSE:

	ExtndDrvKdPrint (("EXTNDDRV.SYS: IRP_MJ_CLOSE\n"));

	break;
    }
    ntStatus = Irp->IoStatus.Status;

    IoCompleteRequest (Irp,
		       IO_NO_INCREMENT
		       );

    return ntStatus;
}



VOID
ExtndDrvUnload(
    IN PDRIVER_OBJECT DriverObject
    )
{
    WCHAR                  deviceLinkBuffer[]  = L"\\DosDevices\\EXTNDDRV";
    UNICODE_STRING         deviceLinkUnicodeString;


    RtlInitUnicodeString (&deviceLinkUnicodeString,
			  deviceLinkBuffer
			  );

    IoDeleteSymbolicLink (&deviceLinkUnicodeString);


    IoDeleteDevice (DriverObject->DeviceObject);

    ExtndDrvKdPrint (("EXTNDDRV.SYS: unloading\n"));
}
